
/*
 *	nntp.c
 *
 *	QNX 4
 *
 *	(C) Copyright 1995 by Robert Krten, all rights reserved.
 *
 *	This module conforms to RFC-977, and performs NNTP functions.
 *
 *	1995 12 21	R. Krten		created
 *	1996 03 12	R. Krten		cleaned up for release
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include "newslib.h"

extern	char	*progname;					/* main.c */
extern	char	*version;					/* version.c */

extern	int		optv;						/* main.c */

/*
   LIST

   Returns a list of valid newsgroups and associated information.  Each
   newsgroup is sent as a line of text in the following format:

      group last first p

   where <group> is the name of the newsgroup, <last> is the number of
   the last known article currently in that newsgroup, <first> is the
   number of the first article currently in the newsgroup, and <p> is
   either 'y' or 'n' indicating whether posting to this newsgroup is
   allowed ('y') or prohibited ('n').

   The <first> and <last> fields will always be numeric.  They may have
   leading zeros.  If the <last> field evaluates to less than the
   <first> field, there are no articles currently on file in the
   newsgroup.

   215 list of newsgroups follows
*/

void
nntp_LIST (reader, startEnd)
void	(*reader) (char *);
void	(*startEnd) (int);
{
	char	buf [BUFSIZ];

	sockwrite ("LIST");
	sockread (buf);
	if (*buf != '2') {
		return;
	}
	genericDotReadDriver (buf, reader, startEnd);
}

/*
   NEWNEWS newsgroups date time [GMT] [<distribution>]

   A list of message-ids of articles posted or received to the specified
   newsgroup since "date" will be listed. The format of the listing will
   be one message-id per line, as though text were being sent.  A single
   line consisting solely of one period followed by CR-LF will terminate
   the list.

   Date and time are in the same format as the NEWGROUPS command.

   A newsgroup name containing a "*" (an asterisk) may be specified to
   broaden the article search to some or all newsgroups.  The asterisk
   will be extended to match any part of a newsgroup name (e.g.,
   net.micro* will match net.micro.wombat, net.micro.apple, etc). Thus
   if only an asterisk is given as the newsgroup name, all newsgroups
   will be searched for new news.

   (Please note that the asterisk "*" expansion is a general
   replacement; in particular, the specification of e.g., net.*.unix
   should be correctly expanded to embrace names such as net.wombat.unix
   and net.whocares.unix.)

   Conversely, if no asterisk appears in a given newsgroup name, only
   the specified newsgroup will be searched for new articles. Newsgroup
   names must be chosen from those returned in the listing of available
   groups.  Multiple newsgroup names (including a "*") may be specified
   in this command, separated by a comma.  No comma shall appear after
   the last newsgroup in the list.  [Implementors are cautioned to keep
   the 512 character command length limit in mind.]

   The exclamation point ("!") may be used to negate a match. This can
   be used to selectively omit certain newsgroups from an otherwise
   larger list.  For example, a newsgroups specification of
   "net.*,mod.*,!mod.map.*" would specify that all net.<anything> and
   all mod.<anything> EXCEPT mod.map.<anything> newsgroup names would be
   matched.  If used, the exclamation point must appear as the first
   character of the given newsgroup name or pattern.

   The optional parameter "distributions" is a list of distribution
   groups, enclosed in angle brackets.  If specified, the distribution
   portion of an article's newsgroup (e.g, 'net' in 'net.wombat') will
   be examined for a match with the distribution categories listed, and
   only those articles which have at least one newsgroup belonging to
   the list of distributions will be listed.  If more than one
   distribution group is to be supplied, they must be separated by
   commas within the angle brackets.

   230 list of new articles by message-id follows

*/

void
nntp_NEWNEWS (since, group, reader, startEnd)
time_t	since;
char	*group;
void	(*reader) (char *);
void	(*startEnd) (int);
{
	struct	tm	*times;
	char	buf [BUFSIZ];

	times = localtime (&since);

	sprintf (buf, "NEWNEWS %s %02d%02d%02d %02d%02d%02d",
			group,
			times -> tm_year > 99 ? times -> tm_year - 100 : times -> tm_year,
			times -> tm_mon + 1,
			times -> tm_mday,
			times -> tm_hour,
			times -> tm_min,
			times -> tm_sec);

	sockwrite (buf);
	sockread (buf);
	if (*buf != '2') {
		return;
	}
	genericDotReadDriver (buf, reader, startEnd);
}

/*
   ARTICLE <message-id>
   HEAD <message-id>
   BODY <message-id>

   Display the header, a blank line, then the body (text) of the
   specified article.  Message-id is the message id of an article as
   shown in that article's header.  It is anticipated that the client
   will obtain the message-id from a list provided by the NEWNEWS
   command, from references contained within another article, or from
   the message-id provided in the response to some other commands.

   Please note that the internally-maintained "current article pointer"
   is NOT ALTERED by this command. This is both to facilitate the
   presentation of articles that may be referenced within an article
   being read, and because of the semantic difficulties of determining
   the proper sequence and membership of an article which may have been
   posted to more than one newsgroup.

   ARTICLE [nnn]

   Displays the header, a blank line, then the body (text) of the
   current or specified article.  The optional parameter nnn is the

   numeric id of an article in the current newsgroup and must be chosen
   from the range of articles provided when the newsgroup was selected.
   If it is omitted, the current article is assumed.

   The internally-maintained "current article pointer" is set by this
   command if a valid article number is specified.

   [the following applies to both forms of the article command.] A
   response indicating the current article number, a message-id string,
   and that text is to follow will be returned.

   The message-id string returned is an identification string contained
   within angle brackets ("<" and ">"), which is derived from the header
   of the article itself.  The Message-ID header line (required by
   RFC850) from the article must be used to supply this information. If
   the message-id header line is missing from the article, a single
   digit "0" (zero) should be supplied within the angle brackets.

   Since the message-id field is unique with each article, it may be
   used by a news reading program to skip duplicate displays of articles
   that have been posted more than once, or to more than one newsgroup.

   220 n <a> article retrieved - head and body follow
           (n = article number, <a> = message-id)
   221 n <a> article retrieved - head follows
   222 n <a> article retrieved - body follows
   223 n <a> article retrieved - request text separately
   412 no newsgroup has been selected
   420 no current article has been selected
   423 no such article number in this group
   430 no such article found
*/

void
nntp_ARTICLE (artid, reader, startEnd)
char	*artid;
void	(*reader) (char *);
void	(*startEnd) (int);
{
	nntp_AHB ("ARTICLE", artid, reader, startEnd);
}

void
nntp_HEAD (artid, reader, startEnd)
char	*artid;
void	(*reader) (char *);
void	(*startEnd) (int);
{
	nntp_AHB ("HEAD", artid, reader, startEnd);
}

void
nntp_BODY (artid, reader, startEnd)
char	*artid;
void	(*reader) (char *);
void	(*startEnd) (int);
{
	nntp_AHB ("BODY", artid, reader, startEnd);
}

void
nntp_AHB (type, artid, reader, startEnd)
char	*type;
char	*artid;
void	(*reader) (char *);
void	(*startEnd) (int);
{
	char	buf [BUFSIZ];

	sprintf (buf, "%s %s", type, artid);
	sockwrite (buf);
	sockread (buf);
	if (*buf != '2') {
		nullDriver (startEnd);
		return;
	}
	genericDotReadDriver (buf, reader, startEnd);
}

void
genericDotReadDriver (buf, reader, startEnd)
char	*buf;
void	(*reader) (char *);
void	(*startEnd) (int);
{
	int		done;

	if (startEnd != NULL) {
		(*startEnd) ('s');		/* indicate Start */
	}
	for (done = 0; !done; ) {
		sockread (buf);
		if (*buf == '.') {
			if (buf [1] == '.') {
				(*reader) (buf + 1);
			}
			if (!buf [1]) {
				done = 1;
			}
		} else {
			(*reader) (buf);
		}
	}
	if (startEnd != NULL) {
		(*startEnd) ('e');		/* indicate End */
	}
}

void
nullDriver (se)
void	(*se) (int);
{
	if (se != NULL) {
		(*se) ('s');
		(*se) ('e');
	}
}

/*
   GROUP ggg

   The required parameter ggg is the name of the newsgroup to be
   selected (e.g. "net.news").  A list of valid newsgroups may be
   obtained from the LIST command.

   The successful selection response will return the article numbers of
   the first and last articles in the group, and an estimate of the
   number of articles on file in the group.  It is not necessary that
   the estimate be correct, although that is helpful; it must only be
   equal to or larger than the actual number of articles on file.  (Some
   implementations will actually count the number of articles on file.
   Others will just subtract first article number from last to get an
   estimate.)

   When a valid group is selected by means of this command, the
   internally maintained "current article pointer" is set to the first
   article in the group.  If an invalid group is specified, the
   previously selected group and article remain selected.  If an empty
   newsgroup is selected, the "current article pointer" is in an
   indeterminate state and should not be used.

   Note that the name of the newsgroup is not case-dependent.  It must
   otherwise match a newsgroup obtained from the LIST command or an
   error will result.

   211 n f l s group selected
           (n = estimated number of articles in group,
           f = first article number in the group,
           l = last article number in the group,
           s = name of the group.)
   411 no such news group
*/

int
nntp_GROUP (group)
char	*group;
{
	char	buf [BUFSIZ];

	sprintf (buf, "GROUP %s", group);
	sockwrite (buf);
	sockread (buf);
	return (*buf == '2');
}

/*
   The first digit of the response broadly indicates the success,
   failure, or progress of the previous command.

      1xx - Informative message
      2xx - Command ok
      3xx - Command ok so far, send the rest of it.
      4xx - Command was correct, but couldn't be performed for
            some reason.
      5xx - Command unimplemented, or incorrect, or a serious
            program error occurred.

   The next digit in the code indicates the function response category.

      x0x - Connection, setup, and miscellaneous messages
      x1x - Newsgroup selection
      x2x - Article selection
      x3x - Distribution functions
      x4x - Posting
      x8x - Nonstandard (private implementation) extensions
      x9x - Debugging output

   In general, 1xx codes may be ignored or displayed as desired;  code
   200 or 201 is sent upon initial connection to the NNTP server
   depending upon posting permission; code 400 will be sent when the
   NNTP server discontinues service (by operator request, for example);
   and 5xx codes indicate that the command could not be performed for
   some unusual reason.

      100 help text
      190
        through
      199 debug output

      200 server ready - posting allowed
      201 server ready - no posting allowed

      400 service discontinued

      500 command not recognized
      501 command syntax error
      502 access restriction or permission denied
      503 program fault - command not performed
*/

