
/*
 *	db_fsys.c
 *
 *	QNX 4
 *
 *	(C) Copyright 1995, 1996 by Robert Krten, all rights reserved.
 *
 *	This module contains all of the database handlers for the VFsys.knews
 *	resource manager.
 *
 *	1995 12 23	R. Krten		created
 *	1996 01 04	R. Krten		updated for VFsys-style
 *	1996 01 15	R. Krten		FileNodeT sorted by expiry
 *	1996 03 12	R. Krten		cleaned up for release
*/

#include "knews.h"

extern	char	*progname;					/* main.c */

extern	int		optD;						/* main.c, debug flag */
extern	int		optv;						/* main.c */

DirNodeT	root;							/* root of the directory tree */
int		createdFlag;						/* global kludge flag, indicates we created a node */

/*
 *	Database management functions
 *
 *	The database consists of a root node (analogous to the root directory
 *	of a filesystem), which contains malloc'd pointers to all of the other
 *	path components.  There are three main data types that are allocated:
 *
 *		Name of this node:
 *			.name, plain char *
 *
 *		Pointers to sibling nodes (across, back), and children/parents (down, up)
 *		nodes:
 *			.across, .back, .down, .up, pointer to DirNodeT
 *
 *		Pointers to file information:
 *			.files, pointer to FileNodeT
*/

void
databaseStarter ()
{
	if (optv > 1) {
		printf ("DBS\n");
		unlink ("debug.out");
	}
	root.name = ".root";		/* should never see this */
	root.across = NULL;
	root.back = NULL;
	root.down = NULL;
	root.up = NULL;
	root.files = NULL;
}

void
databaseEnder ()
{
	if (optD || optv > 1) {
		printf ("DBE\n");
		printf ("Dumping database to debug.out...");
		fflush (stdout);
		debugDumpDatabase ();
		printf ("done\n");
	}

	//	free descendants

	freeDirNode (root.down);
	root.down = NULL;
}

void
freeDirNode (node)
DirNodeT	*node;
{
	DirNodeT	*across;

	while (node != NULL) {
		freeDirNode (node -> down);
		freeFileNode (node -> files);
		if (node -> name != NULL) {
			free (node -> name);
		}
		across = node -> across;
		free (node);
		node = across;
	}
}

void
freeFileNode (node)
FileNodeT	*node;
{
	if (node != NULL) {
		freeFileNode (node -> next);
		free (node);
	}
}

/*
 *	cdMultiLevel / mdMultilevel, and their associated worker
 *	routines, cdOneLevel and mdOneLevel, simulate the operation
 *	of their counterparts under QNX.
*/

DirNodeT *
cdOneLevel (node, path)
DirNodeT	*node;
char		*path;
{

	if (optv > 1) {
		printf ("CD1 (%p %s)\n", node, path);
	}
	for (node = node -> down; node; node = node -> across) {
		if (!strcmp (node -> name, path)) {
			return (node);
		}
	}
	return (NULL);
}

DirNodeT *
cdMultiLevel (path)
char	*path;
{
	DirNodeT		*node;
	StringArrayT	*sa;
	int				i;

	if (optv > 1) {
		printf ("CDM (%s)\n", path);
	}
	node = &root;
	sa = stringToStringArray (path, "./");
	for (i = 0; i < sa -> num && node != NULL; i++) {
		if (optv > 1) {
			printf ("  CDM (%s)\n", sa -> ptrs [i]);
		}
		node = cdOneLevel (node, sa -> ptrs [i]);
	}
	free (sa);
	return (node);
}

DirNodeT *
mdOneLevel (node, path)
DirNodeT	*node;
char		*path;
{
	if (optv > 1) {
		printf ("MD1 (%p %s)\n", node, path);
	}
	if (node -> down == NULL) {
		node -> down = allocDirNode (path);
		if (node -> down) {
			node -> down -> up = node;
		}
		return (node -> down);
	} else {
		for (node = node -> down; node -> across; node = node -> across)
			;
		if (node -> across = allocDirNode (path)) {
			node -> across -> back = node;
		}
		return (node -> across);
	}
}

DirNodeT *
mdMultiLevel (path)
char	*path;
{
	DirNodeT		*new;
	DirNodeT		*node;
	StringArrayT	*sa;
	int				i;

	if (optv > 1) {
		printf ("MDM (%s)\n", path);
	}

	node = &root;
	createdFlag = 0;
	sa = stringToStringArray (path, "./");
	for (i = 0; i < sa -> num && node != NULL; i++) {
		if (optv > 1) {
			printf ("  MDM (%s)\n", sa -> ptrs [i]);
		}
		if ((new = cdOneLevel (node, sa -> ptrs [i])) == NULL) {
			new = mdOneLevel (node, sa -> ptrs [i]);
			createdFlag++;
		}
		node = new;
	}
	free (sa);
	return (node);
}

FileNodeT *
getFileInfo (path)
char	*path;
{
	char		*ptr, *orig;
	FileNodeT	*fnode;

	//	get to the end of the path component

	orig = strdup (path);
	ptr = orig + strlen (orig) - 1;
	while (ptr > orig) {
		if (!isdigit (*ptr)) {
			break;
		}
		ptr--;
	}

	//	if there was no path, (possibly just numbers) fail

	if (ptr == orig) {
		free (orig);
		return (NULL);
	}

	//	convert number, and preterminate path

	*ptr++ = 0;

	fnode = findFileNode (cdMultiLevel (orig), atoi (ptr));
	free (orig);
	return (fnode);
}

FileNodeT *
putFileInfo (path, artnum, expiry, offset, length)
char	*path;
U32		artnum;
U32		expiry;
U32		offset;
U32		length;
{
	DirNodeT	*dnode;
	FileNodeT	*fnode, *new;
	char		buf [BUFSIZ];

	sprintf (buf, "%s/%d", path, artnum);
	if (getFileInfo (buf) != NULL) {
		return (NULL);
	}

	if ((dnode = mdMultiLevel (path)) == NULL) {
		return (NULL);
	}

	if (createdFlag) {
		putGroupInfo (path);
	}

	if ((new = allocFileNode ()) == NULL) {
		return (NULL);
	}

	new -> artnum = artnum;
	new -> expiry = expiry;
	new -> offset = offset;
	new -> length = length;

	if ((fnode = dnode -> files) == NULL) {			// nothing there, add it
		dnode -> files = new;
	} else {										// something there, insert
		if (fnode -> expiry > new -> expiry) {
			new -> next = fnode;
			dnode -> files = new;
		} else {
			while (fnode -> next != NULL
			&&	   fnode -> next -> expiry <= new -> expiry)
			{
				fnode = fnode -> next;
			}
			new -> next = fnode -> next;
			fnode -> next = new;
		}
	}
	return (new);
}
	
FileNodeT *
findFileNode (dnode, num)
DirNodeT	*dnode;
U32			num;
{
	FileNodeT	*fnode;

	if (dnode == NULL || (fnode = dnode -> files) == NULL) {
		return (NULL);
	}

	while (fnode != NULL) {
		if (fnode -> artnum == num) {
			return (fnode);
		}
		fnode = fnode -> next;
	}
	return (fnode);
}

DirNodeT *
allocDirNode (name)
char	*name;
{
	DirNodeT	*new;

	if ((new = calloc (1, sizeof (DirNodeT))) == NULL) {
		return (new);
	}
	new -> name = strdup (name);
	return (new);
}

FileNodeT *
allocFileNode (void)
{
	return (calloc (1, sizeof (FileNodeT)));
}

void
databaseExpire (when)
time_t	when;
{
	databaseExpireDir (&root, when);
}

void
databaseExpireDir (dnode, when)
DirNodeT	*dnode;
time_t	when;
{
	while (dnode) {
		if (dnode -> files) {
			databaseExpireFiles (dnode, when);
		}
		if (dnode -> down) {
			databaseExpireDir (dnode -> down, when);
		}
		dnode = dnode -> across;
	}
}

void
databaseExpireFiles (parentDir, when)
DirNodeT	*parentDir;
time_t	when;
{
	FileNodeT	*tmp;

	while ((parentDir -> files) && (parentDir -> files -> expiry < when)) {
		tmp = parentDir -> files;
		parentDir -> files = parentDir -> files -> next;
		free (tmp);
	}
}

static int debugIndent;

void
debugDumpDatabase ()
{
	FILE	*fp;
	time_t	now;

	if ((fp = fopen ("debug.out", "a")) == NULL) {
		fprintf (stderr, "%s:  unable to open debug.out for a\n", progname);
		perror (NULL);
		exit (1);
	}
	time (&now);
	fprintf (fp, "DEBUG DATABASE DUMP %s", ctime (&now));
	debugIndent = 0;
	debugDumpDirNode (&root, fp);
	fclose (fp);
}

void
debugDumpDirNode (node, fp)
DirNodeT	*node;
FILE		*fp;
{
	char 	buf [32];

	strcpy (buf, "                              ");
	buf [debugIndent] = 0;

	do {
		debugDumpOneDirNode (node, fp, buf);
		if (node -> files) {
			fprintf (fp, "%sFiles (f):\n", buf);
			debugIndent++;
			debugDumpFileNode (node -> files, fp);
			debugIndent--;
		}
		if (node -> down) {
			fprintf (fp, "%sDown (d):\n", buf);
			debugIndent++;
			debugDumpDirNode (node -> down, fp);
			debugIndent--;
		}
		node = node -> across;
	} while (node != NULL);
}

void
debugDumpOneDirNode (node, fp, spaceArray)
DirNodeT	*node;
FILE		*fp;
char		*spaceArray;
{

	fprintf (fp, "%sDN %p \"%s\" -> %p <- %p d %p u %p, f %p:\n",
			spaceArray,
			node,
			node -> name == NULL ? "<NULL>" : node -> name,
			node -> across,
			node -> back,
			node -> down,
			node -> up,
			node -> files);
}

void
debugDumpFileNode (node, fp)
FileNodeT	*node;
FILE		*fp;
{
	char 	buf [32];

	strcpy (buf, "                              ");
	buf [debugIndent] = 0;

	fprintf (fp, "%sFNODE    NEXT     ARTNUM    EXPIRES   OFFSET    LENGTH\n", buf);
	fprintf (fp, "%s-------- -------- --------- --------- --------- ---------\n", buf);
	while (node) {
		fprintf (fp, "%s%p %p %09d %09d %09d %09d\n", buf,
			node,
			node -> next,
			node -> artnum,
			node -> expiry,
			node -> offset,
			node -> length);
		node = node -> next;
	}
}

