
/*
 *	journal.c
 *
 *	QNX 4
 *
 *	(C) Copyright 1996 by Robert Krten, all rights reserved.
 *
 *	This module keeps track of journal entries.  A journal entry
 *	is generated on the following IPC messages:
 *
 *		KNewsMsg_AddArticle
 *
 *	And other interactions occur as a result of these:
 *
 *		KNewsMsg_ExpireArticles
 *		KNewsMsg_EraseDatabase
 *
 *	The journal entries allow us to stop and restart the VFsys.knews
 *	system.  Journal entries are just like news articles -- they have
 *	an expiry, and are stored based upon this expiry.
 *
 *	1996 01 07	R. Krten		created
 *	1996 03 12	R. Krten		cleaned up for release
*/

#include "knews.h"

extern	char	*progname;					/* main.c */
extern	char	*version;					/* version.c */

extern	int		optv;						/* main.c */
extern	ConfigT	*Config_database;			/* main.c, configuration item */

void
journalStarter ()
{
	LPTRT	*lp;
	LPTRT	*orig;
	FILE	*fp;
	time_t	expiry;
	time_t	now;
	int		i;
	char	fname [BUFSIZ];

	/*
	 *	read in the groups file, so that all of the groups that exist
	 *	in the active file are also manifested under the /usr/spool/news
	 *	directory.
	*/

	printf ("groups..."); fflush (stdout);
	sprintf (fname, "%s/groups", Config_database -> items [1].string);
	if ((fp = fopen (fname, "r")) == NULL) {
		if (optv) {
			printf ("no groups file\n");
		}
	} else {
		while (!feof (fp) && fgets (fname, BUFSIZ, fp) != NULL) {
			fname [strlen (fname) - 1] = 0;
			if (optv > 1) {
				printf ("Making newsgroup %s\n", fname);
			}
			mdMultiLevel (fname);
		}
		fclose (fp);
	}

	/*
	 *	now that we have all of the newsgroups that we will be working
	 *	with, go out and read in the unexpired journal files.
	*/

	orig = lptr;

	time (&now);
	journalExpire (now);

	if ((lp = readDir (Config_database -> items [1].string, "*.j")) == NULL) {
		return;
	}

	printf ("journal files (%d)\n", lp -> nlptrs);
	for (i = 0; i < lp -> nlptrs; i++) {
		printf ("%d/%d\r", i, lp -> nlptrs); fflush (stdout);
		sscanf (lp -> object [i], "%d", &expiry);
		sprintf (fname, "%s/%s", Config_database -> items [1].string, lp -> object [i]);
		if (optv > 1) {
			printf ("journal file %s not expired, processing\n", fname);
		}
		processJournalFile (fname);
	}
	lptr = lp;
	lptrInit ('e');
	free (lp);
	lptr = orig;
}

void
journalExpire (when)
time_t	when;
{
	LPTRT	*lp;
	LPTRT	*orig;
	time_t	expiry;
	int		i;
	char	fname [BUFSIZ];

	orig = lptr;

	if ((lp = readDir (Config_database -> items [1].string, "*.j")) == NULL) {
		return;
	}

	for (i = 0; i < lp -> nlptrs; i++) {
		sscanf (lp -> object [i], "%d", &expiry);
		sprintf (fname, "%s/%s", Config_database -> items [1].string, lp -> object [i]);
		if (expiry < when) {
			if (optv > 1) {
				printf ("journal file %s expired, removing\n", fname);
			}
			unlink (fname);
		}
	}
	lptr = lp;
	lptrInit ('e');
	free (lp);
	lptr = orig;
}

void
journalEnder ()
{
}

void
processJournalFile (fname)
char	*fname;
{
	FILE	*fp;
	char	buf [BUFSIZ];
	char	command;
	char	group [BUFSIZ];
	U32		artnum;
	U32		expiry;
	U32		offset;
	U32		length;

	if (optv > 1) {
		printf ("processJournal (%s)...\n", fname);
	}
	if ((fp = fopen (fname, "r")) == NULL) {
		fprintf (stderr, "%s:  couldn't open %s for r\n", progname, fname);
		return;
	}

	while (!feof (fp) && fgets (buf, BUFSIZ, fp) != NULL) {
		switch (*buf) {
		case	'#':		/* comment, ignore */
			break;
		case	'a':		/* add article, parse */
			sscanf (buf, "%c %s %d %d %d %d",
				&command, group, &artnum, &expiry, &offset, &length);

			if (optv > 1) {
				printf ("FJ:  got %c %s %d %d %d %d\n",
						command, group, artnum, expiry, offset, length);
			}

			putFileInfo (group, artnum, expiry, offset, length);
			break;
		}
	}
	fclose (fp);
}

void
writeJournal (expiry, msg)
int		expiry;
char	*msg;
{
	FILE	*fp;
	char	fname [BUFSIZ];
	struct	utimbuf utb;

	sprintf (fname, "%s/%09d.j", Config_database -> items [1].string, expiry);
	if ((fp = fopen (fname, "a")) == NULL) {
		fprintf (stderr, "%s:  can't open journal file %s for a\n", progname, fname);
		fprintf (stderr, "%s:  JOURNAL ENTRIES ARE BEING LOST!\n", progname);
		return;
	}
	fputs (msg, fp);
	fclose (fp);
	utb.actime = utb.modtime = expiry;
	utime (fname, &utb);
}


void
putGroupInfo (grp)
char	*grp;
{
	char	fname [BUFSIZ];
	FILE	*fp;

	sprintf (fname, "%s/groups", Config_database -> items [1].string);
	if ((fp = fopen (fname, "a")) == NULL) {
		fprintf (stderr, "%s:  can't open groups file %s for a\n", progname, fname);
		perror (NULL);
		/* not a super-critical error */
	} else {
		fprintf (fp, "%s\n", grp);
	}
	fclose (fp);
}

