
/*
 *	resmgr.c
 *
 *	QNX 4
 *
 *	(C) Copyright 1995 by Robert Krten, all rights reserved.
 *
 *	This module contains the resource manager.
 *
 *	1995 12 28	R. Krten		created
 *	1996 01 05	R. Krten		updated for VFsys-style
 *	1996 01 08	R. Krten		added IPC debug message
 *	1996 02 18	R. Krten		added lseek/fstat
 *	1996 03 12	R. Krten		cleaned up for release
*/

#include "knews.h"

extern	char	*progname;					/* main.c */
extern	char	*version;					/* version.c */

extern	char	*spooldir;					/* main.c, which dir to manifest in */
extern	int		optp;						/* main.c, priority */
extern	int		optv;						/* main.c, verbosity */

ResmgrT			Resmgr;						/* global */
MsgT			msg;						/* global */

int		io_bogus (pid_t);
int		io_open (pid_t);
int		io_close (pid_t);
int		io_read (pid_t);
int		io_nosys (pid_t);
int		io_lseek (pid_t);
int		io_handle (pid_t);
int		io_fstat (pid_t);
int		io_chdir (pid_t);
int		io_readdir (pid_t);
int		io_rewinddir (pid_t);
int		io_stat (pid_t);

int (*resmgrJT [0x18]) (int) =
{
	io_bogus,			/* 0x0100 */
	io_open,			/* 0x0101 */
	io_close,			/* 0x0102 */
	io_read,			/* 0x0103 */
	io_nosys,			/* 0x0104 iowrite */
	io_lseek,			/* 0x0105 */
	io_nosys,			/* 0x0106 rename */
	io_nosys,			/* 0x0107 get_config */
	io_nosys,			/* 0x0108 dup */
	io_handle,			/* 0x0109 */
	io_fstat,			/* 0x010a */
	io_nosys,			/* 0x010b chmod */
	io_nosys,			/* 0x010c chown */
	io_nosys,			/* 0x010d utime */
	io_nosys,			/* 0x010e flags */
	io_nosys,			/* 0x010f lock */
	io_chdir,			/* 0x0110 */
	io_bogus,			/* 0x0111 */
	io_readdir,			/* 0x0112 */
	io_rewinddir,		/* 0x0113 */
	io_nosys,			/* 0x0114 ioctl */
	io_stat,			/* 0x0115 */
	io_nosys,			/* 0x0116 select */
	io_nosys			/* 0x0117 qioctl */
};

int		newsAddGroup (pid_t);
int		newsAddArticle (pid_t);
int		newsExpire (pid_t);
int		newsEraseDatabase (pid_t);
int		newsWriteDebug (pid_t);

int		(*newsmsgsJT []) (int) =
{
	newsAddGroup,
	newsAddArticle,
	newsExpire,
	newsEraseDatabase,
	newsWriteDebug
};

void
resmgrEnd ()
{
	qnx_prefix_detach (spooldir);
	databaseEnder ();
	exit (0);
}

void
resmgrInit ()
{
	int		i;
	char	buf [BUFSIZ];

	setprio (0, optp);		/* boost our priority */

	/*
	 *	Allow only SIGTERM for graceful termination, ignore all others.
	*/

	for (i = _SIGMIN; i <= _SIGMAX; ++i) {
		if (i != SIGTERM) {
			signal (i, SIG_IGN);
		}
	}

	signal (SIGTERM, &resmgrEnd);	/* setup the terminate signal handler function. */

	/*
	 *	get QNX system messages
	*/

	if (qnx_pflags (~0, _PPF_SERVER, NULL, NULL)) {
		fprintf (stderr, "%s:  couldn't become a server\n", progname);
		perror (NULL);
		exit (1);
	}

	/*
	 *	attach a device number to our I/O manager
	*/

	if ((Resmgr.devno = qnx_device_attach ()) == -1) {
		fprintf (stderr, "%s:  couldn't attach a device\n", progname);
		perror (NULL);
		exit (1);
	}

	/*
	 *	attach a domain of authority
	*/

	if (qnx_prefix_attach (spooldir, NULL, 0)) {
		fprintf (stderr, "%s:  couldn't attach %s\n", progname, spooldir);
		perror (NULL);
		exit (1);
	}

	/*
	 *	for old-style connections and IPC, attach a name
	*/

	strcpy (buf, "/PARSE/VFsys.knews");
	if (qnx_name_attach (0, buf) == -1) {
		fprintf (stderr, "%s:  unable to register name \"%s\", errno %d\n", progname, buf, errno);
		perror (NULL);
		exit (1);
	}

	/*
	 *	funky undocumented magic
	 *
	 *	(This is stolen from QSSL's "iomanager.tar" in /usr/free)
	*/

	if (!(Resmgr.fd_ctrl = __init_fd (getpid ()))) {
		fprintf (stderr, "%s:  couldn't do funky undocumented __init_fd\n", progname);
		perror (NULL);
		exit (1);
	}
}

void
serve ()
{
	pid_t	pid;
	int		replyParms;

	printf ("startup complete.  Server is running\n");
	for (;;) {

		pid = Receive (0, &msg, sizeof (msg));

		if (optv) {
			printEvent (msg.type, pid);
		}

		if (msg.type >= 0x0100 && msg.type <= 0x0117) {
			replyParms = (*resmgrJT [msg.type & 0x00ff]) (pid);
		} else if (msg.type >= MinKNewsMsg && msg.type <= MaxKNewsMsg) {
			replyParms = (*newsmsgsJT [msg.type - MinKNewsMsg]) (pid);
		} else if (msg.type == _SYSMSG) {
			replyParms = sys_msg (pid);
		} else {
			msg.status = ENOSYS;
			replyParms = sizeof (msg.open_reply);
		}

		if (replyParms) {
			Reply (pid, &msg, replyParms);
		}
	}
}

static char *iomsg [] =
{
	"SYS_MSG",		"_IO_OPEN",		"_IO_CLOSE",	"_IO_READ",
	"_IO_WRITE",	"_IO_LSEEK",	"_IO_RENAME",	"_IO_GET_CONFIG",
	"_IO_DUP",		"_IO_HANDLE",	"_IO_FSTAT",	"_IO_CHMOD",
	"_IO_CHOWN",	"_IO_UTIME",	"_IO_FLAGS",	"_IO_LOCK",
	"_IO_CHDIR",	"<bogus>",		"_IO_READDIR",	"_IO_REWINDDIR",
	"_IO_IOCTL",	"_IO_STAT",		"_IO_SELECT",	"_IO_QIOCTL"
};

static char *knmsg [] =
{
	"AddNewsgroup",	"AddArticle",	"Expire",
	"EraseDatabase","WriteDebug"
};

void
printEvent (msgtype, pid)
int		msgtype;
pid_t	pid;
{
	if (msgtype > 0x0100 && msgtype <= 0x0117) {
		printf ("Received I/O message from %d, type %s\n", pid, iomsg [msgtype & 0xff]);
	} else if (msgtype >= MinKNewsMsg && msgtype <= MaxKNewsMsg) {
		printf ("Received KNews message from %d, type %s\n", pid, knmsg [msgtype - MinKNewsMsg]);
	} else if (msgtype == _SYSMSG) {
		printf ("Received SYSMSG message from %d, type %s\n", pid, iomsg [msgtype & 0xff]);
	} else {
		printf ("Received out-of-range message %d (0x%04X) from %d\n", msgtype, msgtype, pid);
	}
}

int
sys_msg (pid)
pid_t	pid;
{
	struct _sysmsg_version_reply ver = { "VFsys.knews", __DATE__, 0, 'A', 0 };
	struct _mxfer_entry mx[2];
	int		vmaj, vmin;

	switch(msg.sysmsg.subtype) {
	case _SYSMSG_SUBTYPE_VERSION:			/* Send driver version back to sin utility. */
		vmaj = (int) (atof (version) * 1000.);
		vmin = vmaj % 10;
		vmaj /= 10;
		ver.version = vmaj;
		ver.letter = vmin + '0';
		msg.sysmsg_reply.status = EOK;
		_setmx(&mx[0], &msg, sizeof(msg.sysmsg_reply));
		_setmx(&mx[1], &ver, sizeof(ver));
		Replymx(pid, 2, &mx);
		return (0);

		/* The following system messages are not supported by this driver. */
	case _SYSMSG_SUBTYPE_DEATH:
	case _SYSMSG_SUBTYPE_SIGNAL:
	case _SYSMSG_SUBTYPE_TRACE:
	default:
		msg.sysmsg_reply.status = ENOSYS;
		return (sizeof (msg.sysmsg_reply));
	}
}

